/* ȿñ㲻ӥåġ
 * shige
 * 02/02 2004
 * sndglis.c (sndhz.c Ȥ)
 *
 * $Id: sndglis.c,v 1.8 2017-03-07 16:48:09+09 shige Exp $
 */
/* sndhz ([option]) [] [Ĺ]
 * sndglis ([option]) [첻] [] [Ĺ]
 * [option]:
 *   -au/wav/aiff: format (default: au)
 *   -ch [n]: channel
 *   -sr [n]: sampling rate
 *   -b [n]: bit length
 *   -v [n]: volume
 *   -o [file]: output file
 *   -wc [s]: ȷ
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "common.h"

#define Maxbuff 8192
#define Minhz 15.0
#define Maxhz 100000.0
#define Maxchannels 8
#define Minsmplrate 100
#define Maxsmplrate 200000

#define Default_volume 0.7
#define Default_basepitch 440.0

#define Default_AU_ch 1
#define Default_WAV_ch 1
#define Default_AIFF_ch 1
#define Default_AU_smpl 8000
#define Default_WAV_smpl 11025
#define Default_AIFF_smpl 11025
#define Default_AU_bit 8 /* fix */
#define Default_WAV_bit 16
#define Default_AIFF_bit 16

char *Commandname[] = { "sndglis", "sndhz", NULL };
enum { ISSNDGLIS = 0, ISSNDHZ, ISUNKNOWN = -1 }; 

int main(int argc, char *argv[]);
double per0(double x);
double per1(double x);
double per2(double x);
double geneglisfunc(int type, double hz1, double hz2, double len,
		    double t);
void putglisdata(FILE *outfile, SNDheader *snd, int format, double vol,
		 int type, double hz1, double hz2, int ch, double smpl,
		 long int blocks);
double genehzfunc(int type, double hz, double t);
void puthzdata(FILE *outfile, SNDheader *snd, int format, double vol,
	       int type, double hz, int ch, double smpl,
	       long int blocks);
double gethz(char c1, char c2, int oct, double base);
int getarg(int argc, char **argv, SNDheader *Snd, int *oform,
	   int *iscmd, double *vol, int *type, double *hz1,
	   double *hz2, int *ch, double *smpl, int *bit, char **lenstr,
	   FILE **outfile);
double str2hz(char *str, double basep);
void usage(int ret, int iscmd);

int main(int argc, char *argv[])
{
    FILE *outfile;
    SNDheader snd;
    double vol, hz1, hz2, flen, smpl;
    int type, ret, format, ch, bit, len, iscmd;
    char *lenstr, buff[Maxbuff];
    long int blocks;
    
    if ((ret = getarg(argc, argv, &snd, &format, &iscmd, &vol, &type,
		      &hz1, &hz2, &ch, &smpl, &bit, &lenstr, &outfile))
	<= 0) {
	usage(ret, iscmd); return -ret;
    }

    /* ch, smpl, bit η */
    switch (format) {
    case E_AU:
	if (ch < 0) ch = Default_AU_ch;
	if (smpl < 0) smpl = Default_AU_smpl;
	if (bit < 0) bit = Default_AU_bit;
	break;
    case E_WAV:
	if (ch < 0) ch = Default_WAV_ch;
	if (smpl < 0) smpl = Default_WAV_smpl;
	if (bit < 0) bit = Default_WAV_bit;
	break;
    case E_AIFF:
	if (ch < 0) ch = Default_AIFF_ch;
	if (smpl < 0) smpl = Default_AIFF_smpl;
	if (bit < 0) bit = Default_AIFF_bit;
	break;
    }
    /* lenstr β */
    if ((len = strlen(lenstr)) <= 0 || len >= Maxbuff) return 50;
    strcpy(buff, lenstr);
    if (buff[len - 1] == 'b') { 
	buff[len - 1] = '\0';
	if ((blocks = atol(buff)) <= 0) return 51;
    } else {
	if ((flen = atof(buff)) <= 0.0) return 52;
	blocks = flen*smpl/1000; 
    }
    /* إåκ */
    setSNDheader(&snd, format, 0, ch, smpl, bit, blocks, 
		 CH_MAGIC | CH_CHANNEL | CH_BITPERSMPL | CH_SMPLRATE
		 | CH_BLOCKS);
    putSNDheader(outfile, &snd, format);
    switch (iscmd) {
    case ISSNDGLIS:
	putglisdata(outfile, &snd, format, vol, type, hz1, hz2, ch, 
		    smpl, blocks);
	break;
    case ISSNDHZ:
	puthzdata(outfile, &snd, format, vol, type, hz1, ch, smpl, 
		  blocks);
	break;
    }
    return 0;
}

/*  1,  -1.0~1.0
 */
double per0(double x)
{
    return sin(2.0*M_PI*x);
}

double per1(double x)
{
    int j;

    j = x;
    return 2.0*(x - j) - 1.0;
}

double per2(double x)
{
    int j;

    j = x;
    return (x - j < 0.5) ? (-1.0) : 1.0;
}

double geneglisfunc(int type, double hz1, double hz2, double len,
		    double t)
{
    double hz, x;
    
    if (hz1 == hz2) hz = hz1*t;
    else { x = hz2/hz1; hz = hz1*(pow(x, t/len) - 1.0)*len/log(x); }
    switch (type) {
    case 0: return per0(hz);
    case 1: return per1(hz);
    case 2: return per2(hz);
    default: return 0.0;
    }
}

void putglisdata(FILE *outfile, SNDheader *snd, int format, double vol,
		 int type, double hz1, double hz2, int ch, double smpl,
		 long int blocks)
{
    long int j;
    int k, bs;
    double x, len;

    bs = SNDblocksize(snd, format);
    if ((len = (double) blocks/smpl) <= 0.0) return;
    for (j=0; j<blocks; j++) {
	x = vol*geneglisfunc(type, hz1, hz2, len, j/smpl);
	for (k=1; k<=ch; k++)
	  SNDput1chleveldata(outfile, x, snd, format);
    }
    if (format != E_AU && bs%2 && blocks%2) fputc(0, outfile);
    return;
}

double genehzfunc(int type, double hz, double t)
{
    switch (type) {
    case 0: return per0(hz*t);
    case 1: return per1(hz*t);
    case 2: return per2(hz*t);
    default: return 0.0;
    }
}

void puthzdata(FILE *outfile, SNDheader *snd, int format, double vol,
	       int type, double hz, int ch, double smpl, long int blocks)
{
    long int j;
    int k, bs;
    double x;

    bs = SNDblocksize(snd, format);
    for (j=0; j<blocks; j++) {
	x = vol*genehzfunc(type, hz, j/smpl);
	for (k=1; k<=ch; k++)
	  SNDput1chleveldata(outfile, x, snd, format);
    }
    if (format != E_AU && bs%2 && blocks%2) fputc(0, outfile);
    return;
}

/* 椯椯ϡΧԥ饹⥵ݡȤ
 */
double gethz(char c1, char c2, int oct, double base)
{
    /* ʿΧܿ */
    double mean[12];
    int tr[] = {9, 11, 0, 2, 4, 5, 7}, d, j;
    double x;

    if (c1 < 'A' || c1 > 'G') return -1.0;
    if (c2 != 0 && c2 != '#' && c2 != 'b') return -2.0;
    mean[0] = 1.0;
    for (j=1; j<12; j++) mean[j] = pow(2.0, (double) j/12.0);

    d = tr[c1 - 'A'];
    if (c2 == '#') {
	if (++ d >= 12) { oct ++; d -= 12; }
    } else if (c2 == 'b') {
	if (-- d < 0) { oct --; d += 12; }
    }
    x = base*mean[d]/mean[9];
    if (oct < 0) for (; oct<0; oct++) x /= 2.0;
    else if (oct > 0) for (; oct>0; oct--) x *= 2.0;
    if (x < Minhz || x > Maxhz) return -3.0;
    else return x;
}

int getarg(int argc, char **argv, SNDheader *Snd, int *oform,
	   int *iscmd, double *vol, int *type, double *hz1,
	   double *hz2, int *ch, double *smpl, int *bit, char **lenstr,
	   FILE **outfile)
{
    char *TypeStr[] = {"sin", "saw", "rect", NULL};
    int j;
    double basep = Default_basepitch;
    int len, len1;

    *outfile = stdout;
    *vol = Default_volume;
    *type = 0;
    *oform = E_AU;
    *ch = -1; 
    *smpl = -1.0;
    *bit = -1;
    *iscmd = ISUNKNOWN;
    *hz2 = -1.0;

    len = strlen(*argv);
    for (j=0; Commandname[j]!=NULL; j++) {
	len1 = strlen(Commandname[j]);
	if (len < len1) continue;
	if (! strcmp(*argv, Commandname[j])) *iscmd = j;
	if (len == len1) continue;
	if (*(*argv + len - len1 - 1) == '/' 
	    && ! strcmp(*argv + len - len1, Commandname[j]))
	  *iscmd = j;
    }
    if (*iscmd == ISUNKNOWN) return -1; /* Unknown command name */

    while (-- argc > 0) {
	if (**(++ argv) == '-') {
	    switch (*(++ *argv)) {
	    case 'h':
	    case 'H':
	    case '?': return 0;
	    case 'o':
		if (*outfile != stdout) return -3;
		if (*(++ *argv) == '\0') {
		    if (-- argc <= 0) return -4; else argv ++; 
		}
		if (! strcmp(*argv, "-")) *outfile = stdout;
		else if ((*outfile = fopen(*argv, "w")) == NULL) {
		    fprintf(stderr, "File [%s] open error.\n", *argv);
		    return -5;
		}
		break;
	    case 'c':
		if (*(++ *argv) == '\0') {
		    if (-- argc <= 0) return -4; else argv ++; 
		}
		if ((*ch = atoi(*argv)) <= 0 || *ch >= Maxchannels) {
		    fprintf(stderr, "Not correct c value (%s).\n", 
			    *argv);
		    return -10;
		}
		break;
	    case 's':
		if (*(++ *argv) == '\0') {
		    if (-- argc <= 0) return -4; else argv ++; 
		}
		if ((*smpl = atof(*argv)) <= Minsmplrate
		    || *smpl >= Maxsmplrate) {
		    fprintf(stderr, "Not correct s value (%s).\n",
			    *argv);
		    return -11;
		}
		break;
	    case 'b':
		if (*(++ *argv) == '\0') {
		    if (-- argc <= 0) return -4; else argv ++; 
		}
		if ((*bit = atoi(*argv)) != 8 && *bit != 16) {
		    fprintf(stderr, "Not correct b value (%s).\n",
			    *argv);
		    return -12;
		}
		break;
	    case 'v':
		if (*(++ *argv) == '\0') {
		    if (-- argc <= 0) return -4; else argv ++; 
		}
		if ((*vol = atof(*argv))< - 1.0 || *vol > 1.0) {
		    fprintf(stderr, "Not correct v value (%s).\n",
			    *argv);
		    return -12;
		}
		break;
	    case 'p':
		if (*(++ *argv) == '\0') {
		    if (-- argc <= 0) return -4; else argv ++; 
		}
		if ((basep = atof(*argv)) <= 0.0) {
		    fprintf(stderr, "Not correct p value (%s).\n",
			    *argv);
		    return -15;
		}
		break;
	    case 't':
		if (*(++ *argv) == '\0') {
		    if (-- argc <= 0) return -4; else argv ++; 
		}
		if (isdigit(**argv)) {
		    if ((*type = **argv - '0') < 0) {
			fprintf(stderr, "Not correct t value (%s).\n",
				*argv);
			return -13;
		    }
		}
		else {
		    for (j=0; TypeStr[j]!=NULL; j++)
		      if (! strcmp(*argv, TypeStr[j])) {
			  *type = j; break; 
		      }
		    if (TypeStr[j] == NULL) {
			fprintf(stderr, "Not correct t value (%s).\n",
				*argv);
			return -13;
		    } 
		}
		break;
	    case 'a':
	    case 'w':
		if (! strcmp(*argv, "au")) *oform = E_AU;
		else if (! strcmp(*argv, "wav")) *oform = E_WAV;
		else if (! strcmp(*argv, "aiff")) *oform = E_AIFF;
		else {
		    fprintf(stderr, "Unknown option (%s)\n", *argv);
		    return -14;
		}
	    case '\0': break;
	    default: return -6;
	    }
	    if (**argv == '\0') { -- *argv; break; }
	} else break;
    }
    switch (*iscmd) {
    case ISSNDGLIS: if (argc != 3) return -7; break;
    case ISSNDHZ: if (argc != 2) return -7; break;
    case ISUNKNOWN: return -7;
    }

    /* #####  ##### */
    if ((*hz1 = str2hz(*argv, basep)) < 0.0) return -8;
    argv ++;
    if (*iscmd == ISSNDGLIS) { 
	if ((*hz2 = str2hz(*argv, basep)) < 0.0) return -8;
	argv ++;
    }
    /* ##### Ĺ ##### */
    *lenstr = *argv;
    return 1;
}

double str2hz(char *str, double basep)
{
    double ret;
    char c1, c2;
    int oct;

    if (isdigit(*str)) {
	if ((ret = atof(str)) < Minhz || ret > Maxhz) {
	    fprintf(stderr, "Not correct hz value (%s)\n", str);
	    return -1.0;
	} else return ret;
    } else {
	if ((c1 = *str) < 'A' || c1 > 'G') {
	    fprintf(stderr, "Not correct hz name (%s)\n", str);
	    return -2.0;
	}
	if (*(++ str) == '\0') { oct = 0; c2 = 0; }
	else {
	    if ((c2 = *str) == '#' || c2 == 'b') str ++;
	    else c2 = 0;
	    if (*str == '\0') oct = 0;
	    else oct = atoi(str);
	}
	if ((ret = gethz(c1, c2, oct, basep)) <= 0.0) {
	    fprintf(stderr, "Not correct hz name\n");
	    return -3.0;
	} else return ret;
    }
}

void usage(int ret, int iscmd)
{
    char *mess[] = {
	"  [length]: Ĺ (ߥ or ֥åĹ)",
	"  [option]:",
	"    -au/wav/aiff: ǡ (default: AU)",
	"    -c[n]: channel ",
	"    -s[n]: ץ󥰼ȿ",
	"    -b[n]: ̻Ҳӥåȿ (8 or 16)",
	"    -v[n]: ܥ塼 (-1.0~1.0; default: 0.7)",
	"    -t[s]: ȷ (default: )",
	"    -o[file]: ϥեλ (default: ɸ)",
	"    -p[n]: ܥԥå (A0 μȿ; default 440Hz)",
	" ",
	" [1] ϥե '-' ꤷ硢ޤϲꤷʤ",
	"     ɸϤˤʤ롣",
	" [2] ϡȿ (μ¿)ޤϲ̾ǻꤹ롣̾",
	"       [ե٥å](#|b)([])",
	"     ηǻꤹ롣줾졢",
	"       [ե٥å]: C(),D(),E(),F(ե),G(),A(),B()",
	"       #: Ⱦ夲롢b: Ⱦ (ά)",
	"       []: ֻ (,...,-3,-2,-1,0,1,2,3,...; ά)",
	"     ֿ̣礭⤤֤̣ά",
	"      0 Ȥ롣㤨СF#2, Cb, D-1 Τ褦˻ꤹ롣",
	"     ȼȿδطϡC0-C1 Ǥϰʲ̤ (A0 = 440Hz ξ)",
	"       C0: 261.6256Hz, C#0(= Db0): 277.1826Hz,",
	"       D0: 293.6648Hz, D#0(= Eb0): 311.1270Hz,",
	"       E0: 329.6276Hz,",
	"       F0: 349.2282Hz, F#0(= Gb0): 369.9944Hz,",
	"       G0: 391.9954Hz, G#0(= Ab0): 415.3047Hz,",
	"       A0: 440.0000Hz, A#0(= Bb0): 466.1638Hz,",
	"       B0: 493.8833Hz,",
	"       C1: 523.2511Hz,",
	"       (ʿΧ)",
	"      1 ־夬 ()  2  (Ⱦʬ) ˤʤ롣",
	" [3] ̾ϡCb1 = B0, B#0 = C1, E#0 = F0, Fb0 = E0 Ȥ롣",
	" [4] ץ󥰼ȿΥǥեȤοͤϡǡ˰ۤʤ롣",
	"           |channel|sampling|bit ",
	"       ----+-------+--------+------",
	"       AU  |  1ch  |  8000Hz|  8bit",
	"       WAV |  2ch  | 11025Hz| 16bit",
	"       AIFF|  2ch  | 11025Hz| 16bit",
	" [5] ȷϡߤϼΤΤѰդƤ (ޤ̾)",
	"       -t0/-tsin :  (default)",
	"       -t1/-tsaw : Υ (||)",
	"       -t2/-trect:  (||||)",
	" [6] Ĺϡñʤͤξϥߥä򡢿ͤκǸ 'b' ",
	"      block Ḷ́롣λϤ벻Ĺ",
	"     (block Ĺ)/(ץ󥰼ȿ) [] Ȥʤ롣",
	NULL
    };
    int j;

    switch (iscmd) {
    case ISSNDGLIS:
	fprintf(stderr, 
		"Usage: sndglis ([option]) [height1] [height2] [length]\n");
	fprintf(stderr,
		"  [height1], [height2]:  (ȿ or ̾)\n");
	break;
    case ISSNDHZ:
	fprintf(stderr, "Usage: sndhz ([option]) [height] [length]\n");
	fprintf(stderr,	"  [height]:  (ȿ or ̾)\n");
	break;
    case ISUNKNOWN:
	fprintf(stderr, "Unknown command name. \n");
	return;
    }
    for (j=0; mess[j]!=NULL; j++) fprintf(stderr, "%s\n", mess[j]);
}
